/* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: exceptions.c,v 1.3 2002/02/20 09:42:25 malekith Exp $
 *
 * Runtime environment for gont.
 *
 * Assumes compiler that allows $ in identifiers, such as GCC.
 * (It is required to have Ksi, so this ain't so important)
 */

#include <gont/runtime.h>

/* Exceptions */

/* This could be highly optimized by using contiguous chunk of memory for
 * try blocks, they are always pushed and pop in the right order. 
 * However we leave it for now. 
 */

struct gont_try_block {
	void *longjmpbuf[5];
	void *value;
	struct gont_try_block *chain;
};

static struct gont_try_block *gont_try_chain;

struct gont_try_block *gont_push_try_block(void)
{
	struct gont_try_block *t;

	t = gont_xmalloc(sizeof(*t));
	
	t->value = NULL;
	t->chain = gont_try_chain;
	
	return (gont_try_chain = t);
}

void *gont_pop_raised_value(void)
{
	void *v = gont_try_chain->value;

	gont_try_chain = gont_try_chain->chain;

	return v;
}

void gont_raise(void *v)
{
	gont_try_chain->value = v;
	__builtin_longjmp(gont_try_chain, 1);
	for (;;) {}
}


/*
 * try B1 with e B2 finally B3 -->
 *
 * void *e = null;
 * try:
 *   if (setjmp(push_try_block()))
 *       e = pop_try_block()
 *   else
 *       B1
 * finally:
 *   try:
 *     if (setjmp(push_try_block()))
 *         e = pop_try_block()
 *     else
 *         if (e) switch e B2 with non_default: e = 0
 *   finally:
 *     try:
 *       if (setjmp(push_try_block()))
 *         e = pop_try_block()
 *       else
 *         B3
 *     finally:
 *       if (e) raise(e)
 */
